#include "stdafx.h"


#define UPDATE_INTERVAL .04

bool SimulateRenderingLoad = false;

void InitWindow() {
	glfwInit();
	glfwOpenWindowHint(GLFW_FSAA_SAMPLES, 8);
	glfwOpenWindow(640, 480, 8, 8, 8, 8, 24, 0, GLFW_WINDOW);
	glfwSetWindowPos(300, 200);
	glfwSetWindowTitle("GameLoopSandbox");
	glfwSetTime(0.0);
}

void InitGL() {
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	int w, h;
	glfwGetWindowSize(&w, &h);
	gluPerspective(60.0, (double) w / h, .1, 20.0);
	glViewport(0, 0, w, h);
	glClearColor(0, 0, 0, 0);
}

void ProcessInput() {

	if (glfwGetKey(GLFW_KEY_LSHIFT)) {
		if (!SimulateRenderingLoad) {
			glfwSetWindowTitle("GameLoopSandbox - SIMULATED HEAVY RENDERING LOAD");
		}
		SimulateRenderingLoad = true;
	}
	else {
		if (SimulateRenderingLoad) {
			glfwSetWindowTitle("GameLoopSandbox");
		}
		SimulateRenderingLoad = false;
	}

}

void Cleanup() {
	glfwTerminate();
}


bool UserWantsToExit() {
	return glfwGetKey(GLFW_KEY_ESC) || !glfwGetWindowParam(GLFW_OPENED);
}

double GetTime() {
	return glfwGetTime();
}

void DrawCube() {

	int iterations = SimulateRenderingLoad ? SIMULATED_RENDERING_LOAD_ITERATIONS : 1;
	
	for (int i = 0; i < iterations; ++i) {

		glColor3f(1.0, 1.0, 1.0);

		glBegin(GL_LINES);
			glVertex3f(-1.0, -1.0, -1.0);
			glVertex3f(-1.0,  1.0, -1.0);
			glVertex3f(-1.0,  1.0, -1.0);
			glVertex3f( 1.0,  1.0, -1.0);
			glVertex3f( 1.0,  1.0, -1.0);
			glVertex3f( 1.0, -1.0, -1.0);
			glVertex3f( 1.0, -1.0, -1.0);
			glVertex3f(-1.0, -1.0, -1.0);

			glVertex3f(-1.0, -1.0,  1.0);
			glVertex3f(-1.0,  1.0,  1.0);
			glVertex3f(-1.0,  1.0,  1.0);
			glVertex3f( 1.0,  1.0,  1.0);
			glVertex3f( 1.0,  1.0,  1.0);
			glVertex3f( 1.0, -1.0,  1.0);
			glVertex3f( 1.0, -1.0,  1.0);
			glVertex3f(-1.0, -1.0,  1.0);

			glVertex3f(-1.0, -1.0, -1.0);
			glVertex3f(-1.0, -1.0,  1.0);
			glVertex3f(-1.0,  1.0, -1.0);
			glVertex3f(-1.0,  1.0,  1.0);
			glVertex3f( 1.0,  1.0, -1.0);
			glVertex3f( 1.0,  1.0,  1.0);
			glVertex3f( 1.0, -1.0, -1.0);
			glVertex3f( 1.0, -1.0,  1.0);
		glEnd();
	
	}

}


double rotation = 0.0;
double previousRotation = 0.0;


void Redraw() {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -5.0);
	glRotatef(rotation, .5, 1.0, .2);

	DrawCube();

	glfwSwapBuffers();
}

void Redraw(double interp) {
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	glTranslatef(0.0, 0.0, -5.0);
	glRotatef(previousRotation + interp * (rotation - previousRotation), .5, 1.0, .2);

	DrawCube();

	glfwSwapBuffers();
}


void Update() {
	ProcessInput();
	previousRotation = rotation;
	rotation += 5.0;
}

void Update(double timeElapsed) {
	ProcessInput();
	previousRotation = rotation;
	rotation += 5.0 * timeElapsed / UPDATE_INTERVAL;
}



void GameLoop_Basic() {
	
	while (!UserWantsToExit()) {
		Update();
		Redraw();
	}

}

void GameLoop_ConstantGameSpeedAndFPS() {
	
	double nextUpdateTime = GetTime();

	while (!UserWantsToExit()) {
		Update();
		Redraw();
		nextUpdateTime += UPDATE_INTERVAL;
		double currentTime = GetTime();
		if (nextUpdateTime - currentTime > 0) {
			Sleep(1000 * (nextUpdateTime - currentTime));
		}
	}

}

void GameLoop_VariableGameSpeedMaxFPS() {
	
	double lastUpdateTime = GetTime();
	double currentTime = GetTime();

	while (!UserWantsToExit()) {
		lastUpdateTime = currentTime;
		currentTime = GetTime();
		Update(currentTime - lastUpdateTime);
		Redraw();
	}

}

void GameLoop_ConstantGameSpeedMaxFPS() {
	
	double currentTime = GetTime();
	double nextUpdateTime = GetTime();

	int MaxFrameskip = 10;

	while (!UserWantsToExit()) {
		currentTime = GetTime();

		int substeps = 0;
		while ((currentTime >= nextUpdateTime) && (substeps < MaxFrameskip)) {
			Update();
			++substeps;
			nextUpdateTime += UPDATE_INTERVAL;
		}

		Redraw();
	}

}

void GameLoop_ConstantGameSpeedMaxFPS_Interpolation() {
	
	double currentTime = GetTime();
	double lastUpdateTime = GetTime();
	double nextUpdateTime = GetTime();

	int MaxFrameskip = 10;

	while (!UserWantsToExit()) {
		currentTime = GetTime();
		int substeps = 0;
		while ((GetTime() >= nextUpdateTime) && (substeps < MaxFrameskip)) {
			Update();
			++substeps;
			nextUpdateTime += UPDATE_INTERVAL;
		}
		Redraw((GetTime() + UPDATE_INTERVAL - nextUpdateTime) / UPDATE_INTERVAL);
	}

}

#pragma region MultiThreaded

	double nextUpdateTime = 0.0;

	void GLFWCALL UpdateLoop(void * arg) {
		while (!UserWantsToExit()) {
			double currentTime = GetTime();
			if (currentTime >= nextUpdateTime) {
				Update();
				nextUpdateTime += UPDATE_INTERVAL;
			}
			glfwSleep(0.001);
		}
	}

	void GameLoop_MultiThreaded() {

		GLFWthread t = glfwCreateThread(UpdateLoop, NULL);

		while (!UserWantsToExit()) {
			Redraw((GetTime() + UPDATE_INTERVAL - nextUpdateTime) / UPDATE_INTERVAL);
		}

		glfwWaitThread(t, GLFW_WAIT);

	}

#pragma endregion


void EnterGameLoop() {
	//GameLoop_Basic();
	//GameLoop_ConstantGameSpeedAndFPS();
	//GameLoop_VariableGameSpeedMaxFPS();
	//GameLoop_ConstantGameSpeedMaxFPS();
	GameLoop_ConstantGameSpeedMaxFPS_Interpolation();
	//GameLoop_MultiThreaded();
}

int main(int argc, char* argv[])
{
	InitWindow();
	InitGL();

	EnterGameLoop();

	Cleanup();

	return 0;
}

